"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
require("reflect-metadata");
let __optional = Symbol('optional');
let __keys = Symbol('keys');
let __mapping = Symbol('mapping');
class JSONObject extends Object {
    set(key, value) {
        if (typeof value == 'object') {
            this[key] = new (Reflect.getMetadata("design:type", this, key))(value);
        }
        else {
            let computed = Reflect.getMetadata("design:type", this, key)(value);
            let computed_type = typeof computed;
            let value_type = typeof value;
            if (computed_type == value_type) {
                this[key] = Reflect.getMetadata("design:type", this, key)(value);
            }
            else {
                throw new TypeError(`${this.constructor.name}.${key} requires type '${computed_type}', got '${value_type}' instead`);
            }
        }
    }
    get(key) {
        return this[key];
    }
    constructor(json) {
        super();
        for (let key of JSONObject.jsonKeys(this)) {
            let json_value;
            let json_key;
            // Is there a mapping this value?
            if (Reflect.hasMetadata(__mapping, this, key)) {
                json_key = Reflect.getMetadata(__mapping, this, key);
            }
            else {
                json_key = key;
            }
            json_value = json[json_key];
            if (json_value === undefined) {
                if (this.get(key) === undefined && !JSONObject.isOptional(this, key)) {
                    throw new TypeError(`${this.constructor.name}.${key} is required`);
                }
            }
            else {
                this.set(key, json_value);
            }
        }
    }
    static map(newKey) {
        return Reflect.metadata(__mapping, newKey);
    }
    static isOptional(target, key) {
        return Reflect.getMetadata(__optional, target, key);
    }
    static jsonKeys(target) {
        return Reflect.getMetadata(__keys, target);
    }
    static preprocess(target, key) {
        let optional = this.isOptional(target, key);
        if (optional !== undefined) {
            throw new SyntaxError(`Cannot mark ${target.constructor.name} as required, already set as ${optional ? 'optional' : 'required'}`);
        }
        let keys;
        if (Reflect.hasMetadata(__keys, target)) {
            keys = Reflect.getMetadata(__keys, target);
        }
        else {
            keys = Array();
            Reflect.defineMetadata(__keys, keys, target);
        }
        keys.push(key);
    }
    static required(target, key) {
        JSONObject.preprocess(target, key);
        Reflect.defineMetadata(__optional, false, target, key);
    }
    static optional(target, key) {
        JSONObject.preprocess(target, key);
        Reflect.defineMetadata(__optional, true, target, key);
    }
}
exports.JSONObject = JSONObject;
function required(target, key) {
    JSONObject.required(target, key);
}
exports.required = required;
function optional(target, key) {
    JSONObject.optional(target, key);
}
exports.optional = optional;
function map(jsonKey) {
    return JSONObject.map(jsonKey);
}
exports.map = map;
//# sourceMappingURL=json_object.js.map